from asyncio import get_running_loop

try:
    import js
except ImportError:
    print("You are not currently using JupyterLite environment")


class IndexDB:
    def __init__(self,db_name="trading-strategy-config"):
        self.db=None
        self.open_request = js.indexedDB.open(db_name, 2);
        self.open_request.onsuccess = self.db_success
        self.open_request.onupgradeneeded = self.db_init
        self.open_request.onerror = self.db_error
        self._db_open=get_running_loop().create_future()     

    async def set_file(self,name,content):
        opened=await self._db_open
        if opened:
            t=self.db.transaction("files","readwrite")
            files=t.objectStore("files")
            request=files.put(content,name)
            request.onsuccess=lambda *args:self._request_future.set_result(True) 
            request.onerror=lambda *args:self._request_future.set_result(False)
            self._request_future=get_running_loop().create_future()
            req_success=await self._request_future
            return req_success
        return False
        
    async def get_file(self,name):
        opened=await self._db_open
        if opened:
            t=self.db.transaction("files","readonly")
            files=t.objectStore("files")
            request=files.get(name)
            request.onsuccess=lambda *args:self._request_future.set_result(True) 
            request.onerror=lambda *args:self._request_future.set_result(False)
            self._request_future=get_running_loop().create_future()
            req_success=await self._request_future
            if req_success:
                content=request.result
                return content
        return None

    def db_init(self,*args):
        self.db=self.open_request.result
        self.db.createObjectStore("files")

    def db_success(self,*args):
        self.db=self.open_request.result
        self._db_open.set_result(True)

    def db_error(self,*args):
        self._db_open.set_result(False)



"""This module contains tooling for chain-agnostic address types.

When trading on multiple blockchains, we need to be able to uniquely identify the same smart contract address on multiple chains.

There is an ongoing effort to standard cross-chain addresses.
For more information see the `CAIP project <https://github.com/ChainAgnostic/CAIPs>`_.
"""

from dataclasses import dataclass


class BadChainAddressTuple(Exception):
    """Something was wrong with the constructed chain - address tuple."""
    pass


class InvalidChainId(BadChainAddressTuple):
    """Chain id was not an integer"""


class BadAddress(BadChainAddressTuple):
    """Ethereum address is invalid"""


@dataclass
class ChainAddressTuple:
    """Present one chain-agnostic address."""

    #: See ChainId for more information - here we have just the raw int value
    chain_id: int

    #: Could be checksummed or non-checksummed address
    address: str

    @staticmethod
    def parse_naive(v: str):
        """Parses chain_id and EVM address tuple.

        Example tuple: `1:0xB4e16d0168e52d35CaCD2c6185b44281Ec28C9Dc` - ETH-USDC on Uniswap v2
        """
        assert type(v) == str

        if not v:
            raise BadChainAddressTuple("Empty string passed")

        parts = v.split(":")

        if len(parts) != 2:
            raise BadChainAddressTuple(f"Cannot split chain id in address {v}")

        address = parts[1]
        if not address.startswith("0x"):
            raise BadAddress(f"Does not look like an address: {address}")

        try:
            chain_id = int(parts[0])
        except ValueError:
            raise InvalidChainId(f"Invalid chain_id on {v}")

        if chain_id <= 0:
            raise InvalidChainId("Invalid chain_id")

        return ChainAddressTuple(chain_id, address)





"""OHLCV aggregation function for Pandas


"""
from dataclasses import dataclass

import pandas as pd

from tradingstrategy.direct_feed.timeframe import Timeframe
from tradingstrategy.direct_feed.direct_feed_pair import PairId


@dataclass(slots=True, frozen=True)
class OHLCVCandle:
    """One OHLCV candle in the direct data feed."""

    pair: PairId
    timestamp: pd.Timestamp
    start_block: int
    end_block: int

    open: float
    high: float
    low: float
    close: float
    volume: float
    exchange_rate: float

    @staticmethod
    def get_dataframe_columns() -> dict:
        fields = dict([
            ("pair", "string"),
            ("start_block", "uint64"),
            ("end_block", "uint64"),
            ("open", "float32"),
            ("high", "float32"),
            ("low", "float32"),
            ("close", "float32"),
            ("volume", "float32"),
            ("exchange_rate", "float32"),
        ])
        return fields


def resample_trades_into_ohlcv(
        df: pd.DataFrame,
        timeframe: Timeframe,
) -> pd.DataFrame:
    """Resample incoming "tick" data.

    - The incoming DataFrame is not groupe by pairs yet,
      but presents stream of data from the blockchain.

      It must have columns `pair`, `amount`, `price`, `exchange_rate`.
      It will be indexed by `timestamp` for the operations.

    - There must be a price given externally for the first open

    - Build OHLCV dataframe from individual trades

    - Handle any exchange rate conversion

    - Any missing values will be filled with
      the last valid close

    - Exchange rate resample is the avg exchange rate of the candle.
      Thus, resampled data might not be accurately converted between native quote
      asset and US dollar value. Any currency conversion must happen before.

    ... and be timestamp indexed.

    :param df:
        DataFrame of incoming trades.

        Must have columns `price`, `amount`

    :param freq:

        Pandas frequency string for the new timeframe duration.

        E.g. `D` for daily.

        See https://stackoverflow.com/a/35339226/315168

    :param offset:
        Allows you to "shift" candles

    :param conversion:
        Convert the incomign values if needed

    :return:
        OHLCV dataframes that are grouped by pair.

        Open price is not the price of the previous close,
        but the first trade done within the timeframe.
    """

    assert len(df) > 0, "Empty dataframe"

    df = df.set_index("timestamp")

    df["abs_amount"] = df["amount"].abs()

    df = df.groupby("pair")

    # Calculate high, low and close values using naive
    # resamping. This assumes there were trades within the freq window.
    df2 = df["price"].resample(timeframe.freq).agg({
        'open': 'first',
        'high': 'max',
        'low': 'min',
        'close': 'last'})

    # Retain timestamp as a column as well,
    # because resample() drops it
    df2["timestamp"] = df2.index.get_level_values(1)

    # TODO: Integers here get converted to floats when processed with agg().
    # Figure out how to prevent this.
    blocks_df = df["block_number"].resample(timeframe.freq).agg({
        'end_block': 'max',
        'start_block': 'min'})

    volume_df = df["abs_amount"].resample(timeframe.freq).agg({
        "total_volume": "sum",
        "avg_amount": "mean",
    })

    trade_side_df = df["amount"].resample(timeframe.freq).agg({
        'buys': lambda x: (x > 0).sum(),
        'sells': lambda x: (x < 0).sum()})

    df2["exchange_rate"] = df["exchange_rate"].resample(timeframe.freq).mean()
    df2["start_block"] = blocks_df["start_block"]
    df2["end_block"] = blocks_df["end_block"]
    df2["volume"] = volume_df["total_volume"]
    df2["avg_trade"] = volume_df["avg_amount"]
    df2["buys"] = trade_side_df["buys"]
    df2["sells"] = trade_side_df["sells"]

    # Resample generates NaN values instead of sparse data.
    # In this point, we just drop the rows that have any NaNs in them

    #                        high     low   close  exchange_rate  start_block  end_block
    # pair     timestamp
    # AAVE-ETH 2020-01-01    80.0    80.0    80.0         1600.0          1.0        1.0
    #          2020-01-02    96.0    96.0    96.0         1600.0          2.0        2.0
    # ETH-USDC 2020-01-02  1600.0  1600.0  1600.0            1.0          2.0        2.0
    #          2020-01-03     NaN     NaN     NaN            NaN          NaN        NaN
    #          2020-01-04     NaN     NaN     NaN            NaN          NaN        NaN
    #          2020-01-05  1620.0  1400.0  1400.0            1.0          7.0        8.0

    df2 = df2.dropna()
    return df2


def get_feed_for_pair(df: pd.DataFrame, pair: PairId) -> pd.DataFrame:
    """Get candles for a single pair.

    :param df:
        An OHLCV dataframe.

        Generated with :py:func:`ohlcv_resample_trades`

    :param pair:
        Which pair we are interested in

    :return:
        Dataframe for this pair
    """

    # groupby().resample() produces multi-index
    # where the first index is the pair

    # ipdb> df.index
    # MultiIndex([('AAVE-ETH', '2020-01-01'),
    #             ('AAVE-ETH', '2020-01-02'),
    #             ('ETH-USDC', '2020-01-02'),
    #             ('ETH-USDC', '2020-01-05')],
    #            names=['pair', 'timestamp'])

    # No data, return empty dataframe
    if len(df) == 0:
        return pd.DataFrame()

    # https://stackoverflow.com/a/45563615/315168
    try:
        return df.xs(pair)
    except KeyError as e:
        raise KeyError(f"Could not find pair for address {pair}") from e


def truncate_ohlcv(df: pd.DataFrame, ts: pd.Timestamp) -> pd.DataFrame:
    """Clear the chain tip data to be written again.

    :param ts:
        Drop everything after this (inclusive).
    """
    if len(df) == 0:
        return df
    df = df[df.index.get_level_values('timestamp') < ts]
    return df




"""Reading and consuming datasets."""

import logging
from pathlib import Path
from typing import Optional, List, Tuple

import pyarrow as pa
from pyarrow import parquet as pq, ArrowInvalid

logger = logging.getLogger(__name__)


class BrokenData(Exception):
    """Raised when we cannot read Parquet file for some reason."""

    def __init__(self, msg: str, path: Path):
        super().__init__(msg)
        self.path = path


def read_parquet(path: Path, filters: Optional[List[Tuple]]=None) -> pa.Table:
    """Reads compressed Parquet file of data to memory.

    File or stream can describe :py:class:`tradingstrategy.candle.Candle`
    or :py:class:`tradingstrategy.pair.DEXPair` data.

    Filtering of candle data can be done
    during the read time, so that large dataset files do not need
    to be fully loaded to the memory. This severely reduces
    the RAM usage for low-memory environments when dealing Parquet.

    `For filtering options see Parquet documentation <https://arrow.apache.org/docs/python/generated/pyarrow.parquet.read_table.html>`_.

    See :py:func:`pyarrow.parquet.read_table`.

    :param stream:
        A file input that must support seeking.

    :param filters:
        Parquet read_table filters.

    """
    assert isinstance(path, Path), f"Expected path: {path}"
    f = path.as_posix()
    logger.info("Reading Parquet %s", f)
    # https://arrow.apache.org/docs/python/parquet.html
    try:
        table = pq.read_table(f, filters=filters)
    except ArrowInvalid as e:
        raise BrokenData(f"Could not read Parquet file: {f}\n"
                         f"Probably a corrupted download.\n"
                         f"See https://tradingstrategy.ai/docs/programming/troubleshooting.html#resetting-the-download-cache\n"
                         f"for instructions how to clear download cache and remove corrupted files, or try the command line:\n"
                         f"rm '{f}'",
                         path=path) \
                        from e
    return table



"""Old Backtrader integration.

.. warning::

    Deprecased. Do not use anymore. Use `trade-executor` framewor instead.
"""

import logging
import datetime
from typing import Iterable, List, Optional, Union, Set

import backtrader as bt
from backtrader import Trade, LineIterator, TradeHistory, num2date, BuyOrder, SellOrder, AutoOrderedDict
from backtrader.feeds import PandasData
import pandas as pd

from tradingstrategy.analysis.tradeanalyzer import TradeAnalyzer, AssetTradeHistory, SpotTrade
from tradingstrategy.timebucket import TimeBucket
from tradingstrategy.pair import DEXPair, PandasPairUniverse


logger = logging.getLogger(__name__)


def convert_backtrader_timestamp(bt_time: float) -> pd.Timestamp:
    """Convert traderader internal timestamps to Pandas.

    Backtrader internally tracks time as "float of day numbers."
    Go figures?

    See :py:meth:`PandasData._load` for more information.

    See :py:func:`num2date` for more information.
    """
    dt = num2date(bt_time)
    return pd.Timestamp(dt)


class DEXFeed(PandasData):
    """A Pandas data feed with token metadata added.

    This feed serves Backtrader Cerebro engine. It contains only raw OHLVC info.
    """

    def __init__(self, pair_info: DEXPair):
        self.pair_info = pair_info
        super(DEXFeed, self).__init__()


class DEXStrategy(bt.Strategy):
    """A strategy base class with support for Trading Strategy DEX specific use cases."""

    def __init__(self, *args, **kwargs):
        super(DEXStrategy, self).__init__(*args, **kwargs)

        #: Currently open position
        self.last_opened_buy: Optional[BuyOrder] = None

        #: The next() tick counter
        self.tick: Optional[int] = None

        self._tradehistoryon = True

    def buy(self, *args, **kwargs) -> BuyOrder:
        """Stamps each trade with a timestamp.

        Normal Backtrader does not have this functionality.
        """

        assert "hint" in kwargs, "Please give hint parameter to buy()"

        buy: BuyOrder = super().buy(*args, **kwargs)
        # Save the trade for the stop loss management
        self.last_opened_buy = buy

        # Hint will be available as order.info.hint
        return buy

    def close(self, *args, **kwargs):
        ret = super().close(*args, **kwargs)
        self.last_opened_buy = None
        return ret

    def get_timestamp(self) -> pd.Timestamp:
        """Get the timestamp of the current candle"""
        bt_time = self.datetime[0]
        return convert_backtrader_timestamp(bt_time)

    def _start(self):
        """"Add tick counter"""
        super()._start()
        self.tick = 0

    def _oncepost(self, dt):
        """Add tick counter."""

        for indicator in self._lineiterators[LineIterator.IndType]:
            if len(indicator._clock) > len(indicator):
                indicator.advance()

        if self._oldsync:
            # Strategy has not been reset, the line is there
            self.advance()
        else:
            # strategy has been reset to beginning. advance step by step
            self.forward()

        self.lines.datetime[0] = dt
        self._notify()

        minperstatus = self._getminperstatus()
        if minperstatus < 0:
            self.tick += 1
            self.next()
        elif minperstatus == 0:
            self.nextstart()  # only called for the 1st value
        else:
            self.prenext()

        self._next_analyzers(minperstatus, once=True)
        self._next_observers(minperstatus, once=True)

        self.clear()


def prepare_candles_for_backtrader(candles: pd.DataFrame) -> pd.DataFrame:
    """Prepare DataFrame format so that Backtrader strategy can read it.

    What assumptions :py:meth:`Celebro.addfeed` makes about Pandas data.
    """

    # Our index must be the timestamp
    candles = candles.set_index(pd.DatetimeIndex(candles["timestamp"]))

    # Create volume column
    candles["volume"] = candles["buy_volume"] + candles["sell_volume"]

    return candles


def reindex_pandas_for_backtrader(df: pd.DataFrame, start: pd.Timestamp, end: pd.Timestamp, bucket: TimeBucket):
    """Backtrader does not allow sparse data, but all data must be filled.

    Sparse data: not trades at the beginning, end of the time series. Missed trading days.
    """

    freq = bucket.to_frequency()
    idx = pd.date_range(start, end, freq=freq)

    # https://stackoverflow.com/a/19324591/315168
    # https://stackoverflow.com/questions/47231496/pandas-fill-missing-dates-in-time-series
    # df.index = idx
    # Backtrader only cares about OHLCV values,
    # so we set everything to zero on missing days
    # TODO: Copy previous day open/close/high/etc here.
    # df.index = pd.DatetimeIndex(df.index)
    reindexed = df.reindex(idx, fill_value=0)
    return reindexed


def add_dataframes_as_feeds(
        cerebro: bt.Cerebro,
        pair_universe: PandasPairUniverse,
        datas: Iterable[pd.DataFrame],
        start: datetime.datetime,
        end: datetime.datetime,
        bucket: TimeBucket,
        plot=False):
    """Add Pandas candle data as source feed to Backtrader strategy tester.

    For each py:class:`pd.DataFrame` creates a new :py:meth:`bt.Celebro.adddata` feed
    of the type :py:class:`CapitalgramFeed`.
    Data on any missing dates is gracefully handled.

    :param plot: Whether Backtrader includes this series in its default plot
    """

    datas = list(datas)

    # TODO HAX
    # Backtrader cannot iterate over broken data if some feeds have data and some do not.
    # With unluck Backtrader stops at the first empty day of a random feed.
    # We mitigate this by assuming the longest feed has all days and
    # longest feed is the first
    datas = sorted(datas, key=lambda df: len(df), reverse=True)

    for df in datas:
        pair_id = df["pair_id"][0]
        pair_data = pair_universe.get_pair_by_id(pair_id)

        # Drop unnecessary columns
        df = df[["open", "high", "low", "close", "volume"]]

        # Reindex so that backtrader can read data
        df = reindex_pandas_for_backtrader(df, start, end, bucket)

        backtrader_feed = DEXFeed(pair_data, dataname=df, plot=plot)
        cerebro.adddata(backtrader_feed)


def analyse_strategy_trades(trades: List[Trade]) -> TradeAnalyzer:
    """Build a trade analyzer from Backtrader executed portfolio."""
    histories = {}

    trade_id = 1

    # Each Backtrader Trade instance presents a position
    # Trade instances contain TradeHistory entries that present change to this position
    # with Order instances attached
    for t in trades:

        assert t.historyon, "Trade history must be on in Backtrader to analyse trades"

        feed: DEXFeed = t.data
        pair_info = feed.pair_info
        pair_id = pair_info.pair_id
        assert type(pair_id) == int

        histentry: TradeHistory
        for histentry in t.history:

            logger.debug("Processing history entry %s: %s", t.tradeid, histentry)

            history = histories.get(pair_id)
            if not history:
                history = histories[pair_id] = AssetTradeHistory()

            status = histentry.status

            assert status.status in (Trade.Open, Trade.Closed), f"Got status {status}"

            event: AutoOrderedDict = histentry.event
            order: Union[BuyOrder, SellOrder] = event.order

            # Internally negative quantities are for sells
            quantity = event.size
            timestamp = convert_backtrader_timestamp(status.dt)
            price = event.price or status.price or order.price or 0
            # price = order.price
            hint = histentry.event.order.info.hint

            # print("Got event", event, status)
            assert quantity != 0, f"Got bad quantity for {status}"
            # import ipdb ; ipdb.set_trace()
            assert price > 0, f"Got invalid trade event {event}, status {status}, order {order}"

            trade = SpotTrade(
                pair_id=pair_id,
                trade_id=trade_id,
                timestamp=timestamp,
                price=price,
                quantity=quantity,
                commission=event.commission,
                slippage=0,  # TODO
                hint=hint,
            )
            history.add_trade(trade)
            trade_id += 1

    return TradeAnalyzer(asset_histories=histories)


class TradeRecorder(bt.Analyzer):
    """Record all trades during the backtest run so that they can be analysed."""

    def create_analysis(self):
        self.trades: List[Trade] = []
        self.existing_trades: Set[int] = set()

    def stop(self):
        pass

    def notify_trade(self, trade: Trade):
        assert isinstance(trade, Trade)
        if trade.ref not in self.existing_trades:
            # One trade got multiple notifications,
            # see if start tracking a new trade
            self.trades.append(trade)
            self.existing_trades.add(trade.ref)

    def get_analysis(self) -> dict:
         # Internally called by Backtrader
         return {"trades": self.trades}




"""Stablecoin supported functions."""

#: Token symbols that are stablecoin like.
#: Note that it is *not* safe to to check the token symbol to know if a token is a specific stablecoin,
#: but you always need to check the contract address.
#: Checking against this list only works
STABLECOIN_LIKE = ("DAI", "USDC", "USDT", "DAI", "BUSD", "UST", "USDN", "LUSD", "VUSD", "USDV", "EUROe", "EURT", "USDP", "iUSD", "USDS", "gmUSD", "USDR", "RAI", "EURS", "TUSD", "EURe", "USD+", "EUROC", "USDs", "USDT.e", "USDC.e")

#: Stablecoins plus their interest wrapped counterparts on Compound and Aave.
#: Also contains other derivates.
WRAPPED_STABLECOIN_LIKE = ("cUSDC", "cUSDT", "sUSD", "aDAI", "cDAI", "tfUSDC", "alUSD", "agEUR", "gmdUSDC", "gDAI")

#: All stablecoin likes - both interested bearing and non interest bearing.
ALL_STABLECOIN_LIKE = STABLECOIN_LIKE + WRAPPED_STABLECOIN_LIKE


def is_stablecoin_like(token_symbol: str, symbol_list=ALL_STABLECOIN_LIKE) -> bool:
    """Check if specific token symbol is likely a stablecoin.

    Useful for quickly filtering stable/stable pairs in the pools.
    However, you should never rely on this check alone.

    Note that new stablecoins might be introduced, so this check
    is never going to be future proof.

    :param token_symbol:
        Token symbol as it is written on the contract.
        May contain lower and uppercase latter.

    :param symbol_list:
        Which filtering list we use.
    """
    assert isinstance(token_symbol, str), f"We got {token_symbol}"
    return (token_symbol.upper() in symbol_list)


"""Trade feed.

Blockchain / exchange agnostic trade feed using Pandas :py:class:`DataFrame` as internal memory buffer.e
"""
import logging
from abc import abstractmethod
from dataclasses import dataclass, asdict
from decimal import Decimal
from typing import Dict, Optional, List, Iterable, Type, TypeAlias, Protocol

import pandas as pd
from numpy import isnan
from tqdm import tqdm

from eth_defi.price_oracle.oracle import BasePriceOracle
from eth_defi.event_reader.reorganisation_monitor import ReorganisationMonitor, ChainReorganisationResolution

from .direct_feed_pair import PairId
from .timeframe import Timeframe


logger = logging.getLogger(__name__)


#: Hex presentation of transaction hash, such
Binary256: TypeAlias = str


@dataclass(slots=True, frozen=True)
class Trade:
    """Capture information about single trade.

    Designed for technical analysis and trading,
    prices are not intentionally unit accurate and thus
    not suitable for accounting.
    """

    #: Trading pair id.
    #:
    #: Ethereum address. Always lowercased.
    pair: PairId

    #: Block number
    block_number: int

    #: Block hash
    block_hash: str

    #: Block mined at
    timestamp: pd.Timestamp

    #: Block mined at
    tx_hash: Binary256

    #: Trade index within the block
    log_index: int

    #: Trade price in quote token
    price: Decimal

    #: Trade amount in quote token.
    #:
    #: Positive for buys, negative for sells.
    amount: Decimal

    #: The US dollar conversion rate.
    #:
    #: You can use this to convert price and amount to US dollars.
    exchange_rate: Decimal

    def __repr__(self):
        return f"<Trade pair: {self.pair}, price: {self.price}, amount: {self.amount}, exchange rate: {self.exchange_rate}>"

    def __post_init__(self):
        assert self.timestamp.tzinfo is None, "Don't use timezone aware timestamps - everything must be naive UTC"
        assert self.tx_hash, "tx_hash missing"

    @staticmethod
    def get_dataframe_columns() -> dict:
        fields = dict([
            ("pair", "string"),
            ("block_number", "uint64"),
            ("block_hash", "string"),
            ("timestamp", "datetime64[s]"),
            ("tx_hash", "string"),
            ("log_index", "uint32"),
            ("price", "object"),
            ("amount", "object"),
            ("exchange_rate", "object"),
        ])
        return fields

    @staticmethod
    def filter_buys(df: pd.DataFrame) -> pd.DataFrame:
        """Filter buy in trades."""
        return df.loc[df.amount > 0]

    @staticmethod
    def filter_sells(df: pd.DataFrame) -> pd.DataFrame:
        """Filter buy in trades."""
        return df.loc[df.amount < 0]


@dataclass(slots=True, frozen=True)
class TradeDelta:
    """Trade history delta for a snapshot"""

    #: Running counter for each connected client
    cycle: int

    #: New incoming data starting this block.
    #:
    #: The client might have already data older than > start_block
    #: from previous delta. In this case the client needs to
    #: purhe  this data as it is discarded due to chain reorganisation.
    start_block: int

    # The start block we are using if we do not do timeframe snap.
    #
    unadjusted_start_block: int

    #: Last block for which we have data
    #:
    end_block: int

    #: Timestamp of the start block
    #:
    start_ts: pd.Timestamp

    #: Timestamp of the end block
    #:
    end_ts: pd.Timestamp

    #: Did we detect any chain reorgs on this update cycle?
    #:
    reorg_detected: bool

    #: List of individual trades
    #:
    #: This dataframe is snapped to the given time frame to make candle processing easier,
    #: and may contain
    #: historical data from the previous cycles.
    trades: pd.DataFrame

    #: List of trades added in this batch
    #:
    new_trades: pd.DataFrame


class SaveHook(Protocol):
    """Save hook called by trade feed when downloading data.

    Called every N seconds.
    """

    def __call__(self, feed: "TradeFeed"):
        pass


class TradeFeed:
    """Real-time blockchain trade feed.

    In-memory latency optimised OHLCv producer for on-chain trades.

    - Keep events in RAM

    - Generate candles based on events

    - Gracefully handle chain reorganisations and drop
      stale data
    """

    def __init__(self,
                 pairs: List[PairId],
                 oracles: Dict[PairId, BasePriceOracle],
                 reorg_mon: ReorganisationMonitor,
                 timeframe: Timeframe = Timeframe("1min"),
                 data_retention_time: Optional[pd.Timedelta] = None,
                 save_hook: Optional[SaveHook] = None,
                 ):
        """
        Create new real-time OHLCV tracker.

        :param pairs:
            List of pool addresses, or list of similar identifiers.

        :param oracles:
            Reference prices for converting ETH or other crypto quoted
            prices to US dollars.

            In the form of pair -> Price oracle maps.

        :param reorg_mon:
            Reorganisation detector and last good block state for a chain.

        :param timeframe:
            Expected timeframe of the data.
            Any trade deltas are snapped to the exact timeframe,
            so that when candle data gets updated,
            we always update only full candles and
            never return trade data that would break the timeframe in a middle.

            Default to one minute candles.

        :param data_retention_time:
            Discard entries older than this to avoid
            filling the RAM.

        :param save_hook:
            Sync the downloaded data to disk.

        """

        assert isinstance(reorg_mon, ReorganisationMonitor)

        for o in oracles.values():
            assert isinstance(o, BasePriceOracle)

        for p in pairs:
            assert type(p) == PairId, f"Only ids allowed, got {p}"

        self.pairs = pairs
        self.oracles = oracles
        self.data_retention_time = data_retention_time
        self.reorg_mon = reorg_mon
        self.timeframe = timeframe
        self.cycle = 1
        self.last_save = 0

        self.trades_df = pd.DataFrame()

        # Check that every pair has a exchange rate conversion oracle
        for p in self.pairs:
            assert p in self.oracles, f"Pair {p} lacks price oracle. Add oracles by pair identifiers."

    def __repr__(self):
        if len(self.trades_df) > 0:
            first_ts = self.trades_df.iloc[0]["timestamp"]
            last_ts = self.trades_df.iloc[-1]["timestamp"]
        else:
            first_ts = last_ts = "-"

        return f"<TradeFeed {first_ts} - {last_ts} with {len(self.trades_df)} trades>"

    def get_block_number_of_last_trade(self) -> Optional[int]:
        """Get the last block number for which we have good data."""

        if len(self.trades_df) == 0:
            return None

        return self.trades_df.iloc[-1]["block_number"]

    def get_trade_count(self) -> int:
        """How many trades we track currently."""
        return len(self.trades_df)

    def add_trades(self, trades: Iterable[Trade], start_block: Optional[int]=None, end_block: Optional[int]=None) -> pd.DataFrame:
        """Add trade to the ring buffer with support for fixing chain reorganisations.

        Transactions may hop between different blocks when the chain tip reorganises,
        getting a new timestamp. In this case, we update the

        .. note::

            It is safe to call this function multiple times for the same event.

        :param start_block:

            Expected block range. Inclusive.

            Used for debug assets

        :param end_block:

            Expected block range. Inclusive.

            Used for debug assets

        :return:
            DataFrame of new trades

        :raise ChainReorganisationDetected:
            If we have detected a block reorganisation
            during importing the data

        """
        data = []

        # For each trade added, check that
        for evt in trades:
            assert isinstance(evt, Trade)
            assert type(evt.block_number) == int, f"Got bad block number {evt.block_number} {type(evt.block_number)}"
            self.reorg_mon.check_block_reorg(evt.block_number, evt.block_hash)
            data.append(asdict(evt))

        new_data = pd.DataFrame(data, columns=list(Trade.get_dataframe_columns().keys()))
        new_data.set_index("block_number", inplace=True, drop=False)

        if len(new_data) > 0:

            if start_block:
                min_block = new_data["block_number"].min()
                if isnan(min_block):
                    logger.error("Bad trade event data:")
                    for idx, r in new_data.iterrows():
                        logger.error("%s: %s", idx, r)
                assert min_block >= start_block, f"Trade event outside desired block range. {min_block} earlier than {start_block}"

            if end_block:
                max_block = new_data["block_number"].max()
                assert max_block <= end_block, f"Trade event outside desired block range. {max_block} later than {end_block}"

        # Check that there is no overlap, any block data should not be duplicated
        if len(self.trades_df) > 0 and len(new_data) > 0:
            last_block_in_buffer = self.trades_df.iloc[-1].block_number
            incoming_block = new_data.iloc[0].block_number
            assert incoming_block > last_block_in_buffer, f"Tried to insert existing data. Last block we have {last_block_in_buffer:,}, incoming data starts with block {incoming_block:,}"

        logger.debug("add_trades(): Existing trades: %s, new trades: %s", len(self.trades_df), len(new_data))

        self.check_duplicates_data_frame(new_data)

        self.trades_df = pd.concat([self.trades_df, new_data])

        return new_data

    def get_latest_trades(self, n: int, pair: Optional[PairId] = None) -> pd.DataFrame:
        """Returns the latest trades.

        These trades will be across all trading pairs we are monitoring.

        :param n:
            Number of trades to return

        :param pair:
            Optional pair to filter with

        :return:
            DataFrame containing n trades.

            See :py:class:`Trade` for column descriptions.

            Return empty DataFrame if no trades.
        """

        if len(self.trades_df) == 0:
            return pd.DataFrame()

        if pair:
            df = self.trades_df.loc[self.trades_df["pair"] == pair]
        else:
            df = self.trades_df
        return df.tail(n)

    def get_latest_price(self, pair: PairId) -> Decimal:
        """Return the latest price of a pair.

        Return the last price record we have.
        """
        df = self.trades_df.loc[self.trades_df["pair"] == pair]
        last = df.iloc[-1]
        return last["price"]

    def truncate_reorganised_data(self, latest_good_block):
        """Discard data because of the chain reorg.

        :param latest_good_block:
            The last block that we cannot discard.
        """

        if len(self.trades_df) > 0:
            self.trades_df = self.trades_df.truncate(after=latest_good_block, copy=False)

    def check_current_trades_for_duplicates(self):
        """Check for duplicate trades.

        Internal sanity check - should not happen.

        Dump debug output to error logger if happens.

        :raise AssertionError:
            Should not happen
        """
        self.check_duplicates_data_frame(self.trades_df)

    def check_enough_history(self,
            required_duration: pd.Timedelta,
            now_: Optional[pd.Timestamp] = None,
            tolerance=1.0,
        ):
        """Check that the dafa we have is good time window wise.

        Internal sanity check - should not happen.

        Dump debug output to error logger if happens.

        :param required_duration:
            How far back we need to have data in our buffer.

        :param now_:
            UTC timestamp what's the current time

        :param tolerance:
            How much tolerance we need.

            Default to 100%, no forgiveness for any lack of data.

            There are several reasons for data mismatch, notably
            being that we estimate blockchain timepoints using block ranges with
            expected block time which is not always stable.

        :raise AssertionError:
            If we do not have old enough data
        """

        if not now_:
            now_ = pd.Timestamp.utcnow().tz_localize(None)

        adjusted_duration = (required_duration * tolerance)
        start_threshold = now_ - adjusted_duration
        first_trade = self.trades_df.iloc[0]
        assert first_trade["timestamp"] <= start_threshold, f"We need data to start at {start_threshold}\n" \
                                                            f"Required duration: {required_duration}, adjusted duration: {adjusted_duration}, now: {now_}, tolerance: {tolerance}, but first trade is:\n{first_trade}"

    @staticmethod
    def check_duplicates_data_frame(df: pd.DataFrame):
        """Check data for duplicate trades.

        - Bugs in the event reader system may cause duplicate trades

        - All trades are uniquely identified by tx_hash and log_index

        - In a perfectly working system duplicate trades do not happen

        :param df:
            Input trades

        :raise AssertionError:
            Should not happen
        """

        # https://stackoverflow.com/questions/59618293/how-to-find-duplicates-in-pandas-dataframe-and-print-them
        duplicate_index = df.duplicated(subset=["tx_hash", "log_index"], keep=False)
        duplicates = df[duplicate_index]

        # Output some hints
        if duplicate_index.any():
            for idx, dup in duplicates.iloc[0:3].iterrows():
                logger.error(f"block: {dup.block_number} {dup.block_hash} {dup.log_index} {dup.amount}")

            unique_blocks = df["block_number"].unique()

            logger.error("Total %d duplicates over %d blocks", len(duplicates), len(unique_blocks))
            raise AssertionError(f"Duplicate trades detected in dataframe")

    @staticmethod
    def check_correct_block_range(df: pd.DataFrame, start_block: int, end_block: int):
        """Check that trades in the given DataFrame are for correct block range.

        - Bugs in the event reader system may cause duplicate trades

        - All trades are uniquely identified by tx_hash and log_index

        - In a perfectly working system duplicate trades do not happen

        :param df:
            Input trades

        :param start_block:
            First block we expect tradse to have. Inclusive.

        :param end_block:
            Last block we expect tradse to have. Inclusive.

        :raise AssertionError:
            Should not happen
        """

        if len(df) == 0:
            return

        first_trade_block = df["block_number"].min()
        last_trade_block = df["block_number"].max()

        assert first_trade_block >= start_block, f"First trade at {first_trade_block:,} was outside the range starting at {start_block:,}"
        assert last_trade_block <= end_block, f"Last trade at {last_trade_block:,} was outside the range ending at {end_block:,}"

    def check_reorganisations_and_purge(self) -> ChainReorganisationResolution:
        """Check if any of block data has changed

        :return:
            Start block since we need to read new data (inclusive).
        """
        reorg_resolution = self.reorg_mon.update_chain()
        self.truncate_reorganised_data(reorg_resolution.latest_block_with_good_data)
        return reorg_resolution

    def backfill_buffer(self, block_count: int, tqdm: Optional[Type[tqdm]] = None, save_hook=None) -> TradeDelta:
        """Populate the backbuffer before starting real-time tracker.

        :param block_count:
            Number of blocks we need to fetch

        :param tqdm:
            A progress visualiser.

            Especially useful during the initial fetch, to show the user how
            long it takes time to fill the buffer.

            Must be `tqdm` context manager compatible.

        :return:
            Data loaded and filled to the work buffer.
        """

        start_block, end_block = self.reorg_mon.load_initial_block_headers(
            block_count=block_count,
            tqdm=tqdm,
            save_callable=save_hook)

        trades = self.fetch_trades(start_block, end_block, tqdm)

        # On initial load, we do not care about reorgs
        return self.update_cycle(start_block, end_block, False, trades)

    def get_exchange_rate(self, pair: str) -> Decimal:
        """Get the current exchange rate for the pair.

        - There is no block number parameter; We always assume
          the data is filled in order, block by block,
          so the exchange rate is always the exchange rate of the current block.

        - Price oracles are updates per-block before
          trade feed for pairs, so price orocles
          can calculate TWAP or similar exchange rate
          for the block.
        """
        return self.oracles[pair].calculate_price()

    def find_first_included_block_in_candle(self, ts: pd.Timestamp) -> Optional[int]:
        """Find the first block number that contains data going into the candle.

        :param ts:
            Timestamp when the block can start (inclusive)

        :return:
            Block number.

            If there is no data, return None.
        """

        after_df = self.trades_df.loc[self.trades_df["timestamp"] >= ts]

        if len(after_df) > 0:
            return after_df.iloc[0]["block_number"]
        else:
            return None

    def update_cycle(self,
                     start_block,
                     end_block,
                     reorg_detected,
                     trades: Iterable[Trade]) -> TradeDelta:
        """Update the internal work buffer.

        - Adds the new trades to the work buffer

        - Updates the cycle number

        - Creates the snapshot of the new trades for the client

        :param start_block:
            Incoming new data, first block (inclusive)

        :param end_block:
            Incoming new data, last  block (inclusive)

        :param reorg_detected:
            Did we detect any chain reorganisations in this cycle

        :param trades:
            Iterable o new trades

        :return:
            Delta of new trades.

            Note that there might not be data for blocks towards `end_block`
            if there were no trades.
        """

        # Update the DataFrame buffer with new trades
        new_trades = self.add_trades(trades, start_block=start_block, end_block=end_block)

        # We need to snap any trade delta update to the edge of candle timeframe
        event_start_ts = self.reorg_mon.get_block_timestamp_as_pandas(start_block)
        data_start_ts = self.timeframe.round_timestamp_down(event_start_ts)
        data_start_block = self.find_first_included_block_in_candle(data_start_ts)

        # If we do not have enough data, then just naively use the given start block
        # and have half-finished candle
        snap_block = data_start_block or start_block

        if len(self.trades_df) > 0:

            # The last block might not contain trades and thus, does not appear in the index.
            # This will cause KeyError in .loc slicing
            last_to_export = min(end_block, self.trades_df.index.values[-1])

            try:
                # Note .loc is inclusive
                # https://medium.com/@curtisringelpeter/understanding-dataframe-selections-and-slices-with-pandas-102a0c2537fb
                # https://medium.com/@curtisringelpeter/understanding-dataframe-selections-and-slices-with-pandas-102a0c2537fb
                exported_trades = self.trades_df.loc[snap_block:last_to_export]
            except KeyError as e:
                # TODO: Figure out
                real_start = self.trades_df.iloc[0]["block_number"]
                real_end = self.trades_df.iloc[-1]["block_number"]
                raise RuntimeError(f"Tried to export trades {snap_block:,} - {last_to_export:,}, but has data {real_start:,} - {real_end:,}") from e
        else:
            exported_trades = pd.DataFrame()

        start_ts = pd.Timestamp.utcfromtimestamp(self.reorg_mon.get_block_timestamp(snap_block)).tz_localize(None)
        end_ts = pd.Timestamp.utcfromtimestamp(self.reorg_mon.get_block_timestamp(snap_block)).tz_localize(None)

        res = TradeDelta(
            self.cycle,
            snap_block,
            start_block,
            end_block,
            start_ts,
            end_ts,
            reorg_detected,
            exported_trades,
            new_trades,
        )

        self.cycle += 1

        return res

    def perform_duty_cycle(self, verbose=False) -> TradeDelta:
        """Update the candle data

        1. Check for block reorganisations

        2. Read new data

        3. Process and index data to candles

        :param verbose:
            More debug logging
        """
        reorg_resolution = self.check_reorganisations_and_purge()
        start_block = reorg_resolution.latest_block_with_good_data + 1
        end_block = self.reorg_mon.get_last_block_read()

        if verbose:
            logger.info(f"Resolved block range to {start_block:,} - {end_block:,}: resolution is {reorg_resolution}")

        # Make sure we do not read ahead of chain tip by accident, because we have forced + 1
        # block above
        if start_block > reorg_resolution.last_live_block:
            start_block = reorg_resolution.last_live_block

        if start_block > end_block:
            # This situation can happen when the lsat block in the chain has reorganised,
            # in such case just read the last block again
            start_block = end_block

        trades = self.fetch_trades(start_block, end_block)

        delta = self.update_cycle(start_block, end_block, reorg_resolution.reorg_detected, trades)
        return delta

    def to_pandas(self, partition_size: int) -> pd.DataFrame:
        df = self.trades_df
        if len(df) > 0:
            df["partition"] = df["block_number"].apply(lambda x: max((x // partition_size) * partition_size, 1))
        return df

    def restore(self, df: pd.DataFrame):
        """Restore data from the previous disk save."""
        logger.debug("Restoring %d trades", len(df))
        self.trades_df = df

    @abstractmethod
    def fetch_trades(self,
                     start_block: int,
                     end_block: Optional[int],
                     tqdm: Optional[Type[tqdm]] = None
                     ) -> Iterable[Trade]:
        """Read data from the chain.

        Add any new trades using :py:meth:`add_trades`.

        :param start_block:
            Start reading from this block (inclusive)

        :param end_block:
            End at this block (inclusive)

        :param tqdm:
            Optional progress bar displayer gadget.

        :raise ChainReorganisationDetected:
            If blockchain detects minor reorganisation during the data ignestion

        :return:
            TradeDelta instance that contains all new trades since start_block.
        """



"""Helpers to ensure timestamp data stays clean."""
import calendar
import datetime

import pandas as pd


def is_compatible_timestamp(ts: pd.Timestamp) -> bool:
    """Ensure Pandas Timestamp is naive.

    We do not carry timezone information in data, because it would slow
    us down.
    """
    assert isinstance(ts, pd.Timestamp), f"We assume pandas.Timestamp, but received {ts}"
    return ts.tz is None


def assert_compatible_timestamp(ts: pd.Timestamp):
    """Check we do not get in bad input timestamps.

    :raise: AssertionError if the timestamp object is incompatible
    """
    assert isinstance(ts, pd.Timestamp), f"not pd.Timestamp: {ts.__class__.__name__}={ts}"
    assert is_compatible_timestamp(ts), f"Received pandas.Timestamp in incompatible format: {type(ts)}: {ts}"


def to_int_unix_timestamp(dt: datetime.datetime) -> int:
    """Get datetime as UTC seconds since epoch."""
    # https://stackoverflow.com/a/5499906/315168
    return int(calendar.timegm(dt.utctimetuple()))
